(************************************************************************)
(* cctasks-sig.sml                                                      *)
(*                                                                      *)
(* Description:                                                         *)
(*     A more simple interface to the ConCert architecture simulator.   *)
(*     See cctasksr-sig.sml for a more complex interface to the         *)
(*     ConCert architecture simulator.                                  *)
(*                                                                      *)
(* Author:                                                              *)
(*     Evan Chang                                                       *)
(************************************************************************)

(* $Id: cctasks-sig.sml,v 1.1.1.1 2002/05/15 22:00:06 bechang_conductor Exp $ *)

(************************************************************************)
(* signature CCTASKS                                                    *)
(*                                                                      *)
(*   Interface:                                                         *)
(*                                                                      *)
(*     This signature refines the interface to the ConCert architecture *)
(*     simulator.  The major differences are that events are not used   *)
(*     in this interface and the select and wrap have been replaced     *)
(*     by a new construct called 'relax' and 'relaxi'.  We believe      *)
(*     relax can be used the replace the common select/wrap idiom       *)
(*     as illustrated below.  The following examples return the         *)
(*     asks for the results from a pair of tasks 't1' and 't2' and      *)
(*     applies 'f' to each result.  Note, 'f' is computed on the        *)
(*     result that received first while waiting for the second result.  *)
(*                                                                      *)
(*       mappair_using_selectwrap : (result -> 'a) -> task * task       *)
(*                                    -> 'a * 'a                        *)
(*       fun mappair_using_selectwrap f (t1,t2) =                       *)
(*             select [                                                 *)
(*               wrap (recvResultEvt t1,                                *)
(*                     fn r1 => (f r1, f (recvResult t2))),             *)
(*               wrap (recvResultEvt t2,                                *)
(*                     fn r2 => let r2' = f r2                          *)
(*                              in (f (recvResult t1), r2') end)        *)
(*             ]                                                        *)
(*                                                                      *)
(*       mappair_using_relaxi : (result -> 'a) -> task * task           *)
(*                                -> 'a * 'a                            *)
(*       fun mappair_using_relaxi f (t1,t2) =                           *)
(*           let                                                        *)
(*	       datatype which = A | B                                   *)
(*           in                                                         *)
(*             case relaxi [(t1,A), (t2,B)] of                          *)
(* 	         ((r1,A), [other]) => (f r1, f (sync other))            *)
(*   	       | ((r2,B), [other]) => let r2' = f r2                    *)
(*                                    in (f (sync other), r2') end      *)
(*           end                                                        *)
(*                                                                      *)
(************************************************************************)
signature CCTASKS =
sig

  type env
  type result

  (* A "result" task *)
  type task
  exception InvalidTask
	    
  (* Start a task  *)
  val inject : bool -> (env -> result) * env -> task

  (* Start a disabled tasks *)
  val enable : task -> unit
  
  (* Get results *)
  val sync : task -> result
  val syncall : task list -> result list
  val relax  : task list -> result * task list
  val relaxi : (task * 'idx) list -> (result * 'idx) * (task * 'idx) list

  (* Indicate a task is no longer needed *)
  val forget : task -> unit

  (* Exit early *)
  val exit : result -> 'a

  (* Query the status of a task *)
  val status : task -> CCStatus.status

  (* Start Simulation *)
  val startSim : real -> (env -> result) * env -> result

end; (* signature CCTASKS *)
