(************************************************************************)
(* cctasks.sml                                                          *)
(*                                                                      *)
(* Description:                                                         *)
(*     Creates a structure that has the simple interface to the ConCert *)
(*     architecture.  See cctasks-sig.sml for a description of          *)
(*     the interface.                                                   *)
(*                                                                      *)
(* Author:                                                              *)
(*     Evan Chang                                                       *)
(************************************************************************)

(* $Id: cctasks.sml,v 1.1.1.1 2002/05/15 22:00:06 bechang_conductor Exp $ *)


(************************************************************************)
(* functor MakeSimple                                                   *)
(*                                                                      *)
(*   Implementation:                                                    *)
(*                                                                      *)
(*     Note that the implementation is parametric in the types 'env'    *)
(*     and 'result' which are obtained from S.                          *)
(*                                                                      *)
(************************************************************************)
functor MakeSimple(S : CCTASKSR)
:> CCTASKS where type result = S.result
             and type env = S.env =
struct

  type result = S.result
  type env = S.env

  type task = S.task
  exception InvalidTask

  (* Start the simulation *)
  val startSim = S.startSim

  (* Query the status of a task *)
  val status = S.status

  (* Indicate a task is no longer needed *)
  val forget = S.forget

  (* Exit early *)
  val exit = S.exit

  (* Start a task  *)
  val inject = S.inject

  (* Start a disabled task *)
  val enable = S.enable
  
  (* Get a result *)
  val sync = S.recvResult

  (* Get each result *)
  val syncall = map sync


  (*************************)
  (* Relax                 *)
  (*************************)
  
  (* buildSelect : 'a event list -> ('a, int) event list *)
  fun buildSelect es =
      let
        fun buildSelect' (nil, _)  = nil
          | buildSelect' (h::t, i) =
              (S.wrap (h, fn x => (x,i)))::(buildSelect' (t, i+1))
      in
        buildSelect' (es, 0)
      end

  (* filteri : int -> 'a list -> 'a option * 'a list *)
  fun filteri i l =
      let
        fun filteri' (nil, _)  k = k nil
          | filteri' (h::t, j) k =
            if i=j then filteri' (t, j+1) (fn l' => (SOME h, #2(k l')))
            else        filteri' (t, j+1) (fn l' => k (h::l'))
      in
        filteri' (l, 0) (fn l' => (NONE, l'))
      end
          
  (* relax  : task list -> result * task list *)
  fun relax nil = raise Fail "empty list in relax"
    | relax l =
      let
        val resultEvts = map S.recvResultEvt l
        val (r,i) = S.select (buildSelect resultEvts)
      in
        (r, #2(filteri i l))
      end

  (* relaxi : (task * 'idx) list -> (result * 'idx) * (task * 'idx) list *)
  fun relaxi nil = raise Fail "empty list in relaxi"
    | relaxi l =
      let
        val (ts,is) = ListPair.unzip l
        val resultEvts = map S.recvResultEvt ts
        val (r,i) = S.select (buildSelect resultEvts)

        val (_, ts')   = filteri i ts
        val (idx, is') = filteri i is
      in
        ((r, valOf idx), ListPair.zip (ts',is'))
      end

end; (* functor MakeSimple *)



(************************************************************************)
(* functor CCTasksFn                                                    *)
(************************************************************************)
functor CCTasksFn(structure T : sig
                                  type env
                                  type result
                                end)
:> CCTASKS where type env = T.env
             and type result = T.result =
struct

  structure CCTR = CCTasksRFn(structure T = T)
  structure S = MakeSimple(CCTR)

  open S

end; (* functor CCTasksFn *)
