(************************************************************************)
(* cctasksr.sml                                                         *)
(*                                                                      *)
(* Description:                                                         *)
(*     Simulate the ConCert architecture using CML.  See                *)
(*     cctasksr-sig.sml for a discussion about the interface, and       *)
(*     see below for a discussion about the implementation.             *)
(*                                                                      *)
(* Author:                                                              *)
(*     Evan Chang                                                       *)
(************************************************************************)

(* $Id: cctasksr.sml,v 1.1.1.1 2002/05/15 22:00:06 bechang_conductor Exp $ *)

(* Uncomment to see uncaught exceptions *)
(* structure TraceCML = TraceCML *)


(************************************************************************)
(* functor CCTasksRFn                                                   *)
(*                                                                      *)
(*   Implementation:                                                    *)
(*                                                                      *)
(*     Note that the implementation is parametric in the types 'env'    *)
(*     and 'result'.                                                    *)
(*                                                                      *)
(*     The simulation relies on two global tables to maintain the       *)
(*     the state of the simulation.  This is clearly a bit unrealistic. *)
(*     This state would really implicitly kept on the network via       *)
(*     some protocol along with local state (work queues, etc).         *)
(*                                                                      *)
(*     ConCert tasks (not ConCert cords) are modeled by CML threads.    *)
(*     This is due to issues with using callcc/throw with CML, which    *)
(*     does make the simulation slightly less realistic but is probably *)
(*     acceptable.  There are two kinds of CML threads running during   *)
(*     the simulation: the threads modeling ConCert tasks and the       *)
(*     threads for misc. bookkeeping (id server, lock servers,          *)
(*     failure simulation, etc.).                                       *)
(*                                                                      *)
(*     Failure simulation is done by a thread that periodically         *)
(*     tells the scheduler to 'forget' about a task, which marks        *)
(*     the status of a task as Failed and issues a FAILED message.      *)
(*     In other words, the simulation informs itself when a "failure"   *)
(*     occurs.  This clearly unrealistic because  it completely         *)
(*     side-steps the issue of failure detection (i.e. when do we       *)
(*     decide a node on the network has fallen off the network?).       *)
(*     Also, we don't model failures other than fail-stop (i.e.         *)
(*     Byzantine failures, drop packets, etc.)                          *)
(*                                                                      *)
(*     The last bit of unrealism is that a task handle is simply        *)
(*     integer being used as an ID rather than using a hash of the      *)
(*     code and the arguments.  This is because we simply can't get     *)
(*     at the code to hash it and is another reason why we cannot       *)
(*     implement the specification as a ML library.                     *)
(*                                                                      *)
(*     Dependencies are roughly modeled by CML ivars, but we are        *)
(*     limited to 1 dependency rather than some conjunction and         *)
(*     disjunction of dependencies.                                     *)
(*                                                                      *)
(************************************************************************)
functor CCTasksRFn(structure T : sig
                                   type env
                                   type result
                                 end)
:> CCTASKSR where type env = T.env
              and type result = T.result =
struct
  open CCStatus

  type env = T.env
  type result = T.result

  (******************************************)
  (* task                                   *)
  (*                                        *)
  (* A task handle is a simply an int that  *)
  (* we use to lookup in the global table.  *)
  (* This at least partially models the     *)
  (* idea that some query process must be   *)
  (* used to find out about the state       *)
  (* of a task.                             *)
  (******************************************)
  type task = int
  exception InvalidTask

  (******************************************)
  (* Msg                                    *)
  (*                                        *)
  (* Msgs are used to communicate results   *)
  (* between "tasks".  FAILURE indicates    *)
  (* that task failed which also carries    *)
  (* with it a wrap function used when      *)
  (* restarting a task.  SUCCESS carries    *)
  (* the result value.                      *)
  (******************************************)
  datatype 'a Msg =
    FAILURE of task * (result -> 'a)
  | SUCCESS of task * 'a

  type 'a event = 'a Msg CML.event


  (******************************************)
  (* Global State                           *)
  (*                                        *)
  (* This is a bit unrealistic.  Any of     *)
  (* this state information must be         *)
  (* maintained via protocols on the        *)
  (* network.                               *)
  (******************************************)

  (*******************************)
  (* Task Handles                *)
  (*******************************)
  (* Generate unique task IDs for each task. *)
  fun createIdSrc () =
      let
        open CML
        val idchan = channel ()

        fun loop nextid =
            let
              val _ = send (idchan, nextid)
            in
              loop (nextid+1)
            end

        (* Start server loop *)
        val _ = spawn (fn () => loop 0)
      in
        fn () => recv idchan
      end

  val newTaskId : (unit -> task) ref = ref (fn () => ~1)


  (*******************************)
  (* Task State                  *)
  (*******************************)
  structure TaskState
  :> sig

      (******************************************)
      (* taskState                              *)
      (*                                        *)
      (* The state of a task is given by the    *)
      (* thread id that models it (if it has    *)
      (* been enabled), the current status,     *)
      (* the code and arguments, and the        *)
      (* communication channel (an ivar) for    *)
      (* the result of a task.  The 'out' field *)
      (* is a result Msg SyncVar.ivar option    *)
      (* because when a task "fails" 'out'      *)
      (* should be NONE so that the garbage     *)
      (* collector will garbage collect the     *)
      (* thread.                                *)
      (******************************************)
      type taskState =
        { tid     : CML.thread_id option,
          status  : status,
          closure : (env -> result) * env,
          out     : result Msg SyncVar.ivar option
        }

       val numTasks : unit -> int

       val get    : task -> taskState
       val add    : task * taskState -> unit
       val update : task * (taskState -> taskState) -> unit
       val updateStatus : task * (status -> status option) -> status option

       val getTaskId : CML.thread_id -> task
       val setTaskId : CML.thread_id * task -> unit

       val beginTaskStart : task -> unit
       val endTaskStart : task -> unit

       val init : unit -> unit
     end =
  struct
    open SyncVar

    type taskState =
      { tid     : CML.thread_id option,          (* id of the thread that
                                                    represents the task *)
        status  : status,                        (* current status *)
        closure : (env -> result) * env,         (* the code and the env *)
        out     : result Msg SyncVar.ivar option (* where to send the ans *)
      }

    (**************************************************)
    (* Task State Table                               *)
    (*                                                *)
    (* Invariants:                                    *)
    (*   + If the task has failed, then the status    *)
    (*     of the task is set to Failed and           *)
    (*     "out" must be NONE.                        *)
    (**************************************************)
    val INIT_TASK_TBLSIZE = 10
    val DEFAULT_ENTRY = NONE
    structure DArr = DynamicArrayFn( MonoArrayFn(struct
                                                   type elem = taskState option
                                                 end) )

    (* tasks : taskState DArr.array *)
    val tasks = DArr.array (INIT_TASK_TBLSIZE, DEFAULT_ENTRY)
                handle e => 
                  let
                    val msg = "Unable to allocate task state table of size "
                              ^ (Int.toString INIT_TASK_TBLSIZE) ^ "!"
                  in
                    (Logger.log Logger.FATAL msg;
                    raise e)
                  end

    val INIT_TASKSTART_TBLSIZE = 10
    val DEFAULT_TASKSTART_ENTRY = NONE
    structure LockArr = DynamicArrayFn( MonoArrayFn(struct
                                                      type elem =
                                                        RWLocks.lock option
                                                    end))

    (* startTaskLocks : RWLocks.lock option LockArr.array *)
    val startTaskLocks = LockArr.array (INIT_TASKSTART_TBLSIZE,
                                        DEFAULT_TASKSTART_ENTRY)
                handle e => 
                  let
                    val msg = "Unable to allocate task state table of size "
                              ^ (Int.toString INIT_TASKSTART_TBLSIZE) ^ "!"
                  in
                    (Logger.log Logger.FATAL msg;
                    raise e)
                  end

    (* nTasks : int ref *)
    val nTasks = ref 0

    (* We have a global lock for the entire table rather than individual
       locks for each entry because the table may resize.  This may be
       a place for optimization *)
    val stateTblLock : RWLocks.lock option ref = ref NONE

    fun bRState () = CML.sync (RWLocks.beginRead (valOf (!stateTblLock)))
    fun eRState () = RWLocks.endRead (valOf (!stateTblLock))
    fun bWState () = CML.sync (RWLocks.beginWrite (valOf (!stateTblLock)))
    fun eWState () = RWLocks.endWrite (valOf (!stateTblLock))

    (**************************************************)
    (* Thread to Task Mapping                         *)
    (**************************************************)
    val INIT_THREADTASK_TBLSIZE = 31
    structure TTTbl =
      HashTableFn(struct
                    type hash_key = CML.thread_id
                     val hashVal  = CML.hashTid
                     val sameKey  = CML.sameTid
                  end)
  
    (* ttTbl : task TTTbl.Table *)
    val ttTbl : task TTTbl.hash_table =
            TTTbl.mkTable (INIT_THREADTASK_TBLSIZE, InvalidTask)
                  handle e =>
                  let
                    val msg = "Unable to allocate thread id to task id table!"
                  in
                    (Logger.log Logger.FATAL msg;
                     raise e)
                  end

    val ttTblLock : RWLocks.lock option ref = ref NONE

    fun bRTT () = CML.sync (RWLocks.beginRead (valOf (!ttTblLock)))
    fun eRTT () = RWLocks.endRead (valOf (!ttTblLock))
    fun bWTT () = CML.sync (RWLocks.beginWrite (valOf (!ttTblLock)))
    fun eWTT () = RWLocks.endWrite (valOf (!ttTblLock))
  
    (**************************************************)
    (* Accessors                                      *)
    (**************************************************)

    (* numTasks : unit -> int *)
    fun numTasks () =
        let
          val _ = bRState ()
          val n = !nTasks
          val _ = eRState ()
        in
          n
        end

    (* get : task -> taskState *)
    fun get taskId =
        let
          val _ = bRState ()
          val SOME e = DArr.sub (tasks, taskId)
          val _ = eRState ()
        in
          e
        end handle _ => (eRState ();
                         raise InvalidTask)

    (* add : task * taskState -> unit *)
    fun add (taskId, e) =
        let
          val _ = bWState ()
          val _ = DArr.update (tasks, taskId, SOME e)
          val _ = LockArr.update (startTaskLocks, taskId,
                                  SOME (RWLocks.lock ()))
          val _ = nTasks := !nTasks + 1
          val _ = eWState ()
        in
          ()
        end handle _ => (eWState ();
                         raise InvalidTask)

    (* update : task * (taskState -> taskState) -> unit *)
    fun update (taskId, f) =
        let
          val _       = bWState ()
          val SOME ts = DArr.sub (tasks, taskId)
          val _       = DArr.update (tasks, taskId, SOME (f ts))
          val _       = eWState ()
        in
          ()
        end handle _ => (eWState ();
                         raise InvalidTask)

    (* updateStatus : task * (Status.status -> Status.status option)
                        -> Status.status option *)
    fun updateStatus (taskId, f) =
        let
          val _       = bWState ()
          val SOME ({tid,status,closure,out}) = DArr.sub (tasks, taskId)
          val res = f status
        in
          ((case res of
             NONE   => ()
           | SOME s => DArr.update (tasks, taskId,
                                    SOME ({tid=tid,status=s,closure=closure,
                                           out=out}))
           );
           eWState ();
           res)
        end handle _ => (eWState ();
                         raise InvalidTask)

    (* getTaskId : CML.thread_id -> task *)
    fun getTaskId threadid = 
        let
          val _ = bRTT ()
          val taskid = TTTbl.lookup ttTbl threadid
          val _ = eRTT ()
        in
          taskid
        end handle _ => (eRTT ();
                         raise InvalidTask)

    (* setTaskId : CML.thread_id * task -> unit *)
    fun setTaskId (threadid, taskid) =
        let
          val _ = bWTT ()
          val _ = TTTbl.insert ttTbl (threadid, taskid)
          val _ = eWTT ()
        in
          ()
        end handle _ => (eWTT ();
                         raise InvalidTask)

    (* beginTaskStart : task -> unit *)
    fun beginTaskStart taskId =
        let
          val _ = bRState ()
          val SOME l = LockArr.sub (startTaskLocks, taskId)
          val _ = eRState ()
        in
          CML.sync (RWLocks.beginWrite l)
        end handle _ => (eRState ();
                         raise InvalidTask)
    
    (* endTaskStart : task -> unit *)
    fun endTaskStart taskId =
        let
          val _ = bRState ()
          val SOME l = LockArr.sub (startTaskLocks, taskId)
          val _ = eRState ()
        in
          RWLocks.endWrite l
        end handle _ => (eRState ();
                         raise InvalidTask)

    (* init : unit -> unit *)
    fun init () =
        (nTasks := 0;
         TTTbl.clear ttTbl;
         stateTblLock := SOME (RWLocks.lock ());
         ttTblLock    := SOME (RWLocks.lock ()))

  end (* structure CCTasksSim.TaskState *)



  (******************************************)
  (* Random Number for Failure Simulation   *)
  (*                                        *)
  (* Invariant:                             *)
  (*    Only one thread will access this;   *)
  (*    otherwise, it is not thread-safe.   *)
  (******************************************)
  local
    val seed =
       let
         val d = Date.fromTimeUniv(Time.now())
         val m = Date.minute d
         val s = Date.second d
       in
         (m,s)
       end

    val rand = Random.rand seed
  in
    (* randTask : unit -> task option *)
    fun randTask () =
       let
          val ntasks = TaskState.numTasks ()
        in
          if ntasks <= 0 then NONE
          else SOME ((Random.randNat rand) mod ntasks)
        end

    (* willFail : real -> bool *)
    fun willFail failureParam = (Random.randReal rand) <= failureParam
  end



  (******************************************)
  (* Task Id                                *)
  (******************************************)

  (* getTaskId : unit -> task *)
  fun getTaskId () =
      let
        val threadid = CML.getTid ()
      in
        TaskState.getTaskId threadid
      end

  (* taskIdToString : task -> string *)
  fun taskIdToString tskId =
        StringCvt.padLeft #"0" 5 (Int.toString tskId)


  (******************************************)
  (* Logging                                *)
  (******************************************)
  (* log : Logger.level -> string -> unit *)
  fun log lvl msg =
      let
        val prefix = (taskIdToString (getTaskId ())) ^ "(" ^
                     (CML.tidToString (CML.getTid ())) ^ "): "
                     handle InvalidTask =>
                       (CML.tidToString (CML.getTid ())) ^ "       : "
      in
        Logger.log lvl (prefix ^ msg ^ "\n")
      end

  val dbgMsg : string -> unit = log Logger.DEBUG
  val noteMsg : string -> unit = log Logger.NOTE

  (******************************************)
  (* Implementation                         *)
  (******************************************)

  (* tValOf : 'a option -> 'a *)
  fun tValOf x = (valOf x) handle Option => raise InvalidTask

  (* status : task -> Status
  
     Return the status of the given task.
  *)
  fun status tskId =
      let
        val {status,...} = TaskState.get tskId
      in
        status
      end

  (* updStat : Status -> Status -> Status option

     Used to only update the status if the status != FAILED
  *)
  fun updStat newStat oldStat =
      (case oldStat of
         Failed => NONE
       | _ => SOME newStat)

  (* startTask : bool -> task -> unit
     
     startTask newIvar tskId = ()
       + If newIvar is true, then create a fresh ivar for this
         tasks result.
  *)
  fun startTask newIvar tskId =
      let
       (* spawnCord : (env -> result) -> env SyncVar.ivar
                        -> result Msg SyncVar.ivar option -> unit *)
       fun spawnCord f inp outp =
           let
             val _ = dbgMsg "Spawning cord ..."

             fun thread () =
                 let
                   fun updTidOut t (ts as {tid,status,closure,out}) =
                         {tid=(SOME t), status=status, closure=closure,
                          out=outp}
                 
                   val myTid = CML.getTid ()
                   val _ = TaskState.setTaskId (myTid, tskId)
                   
                   val _ = (TaskState.update (tskId, updTidOut myTid))
                           handle e => (TaskState.endTaskStart tskId; raise e)

                   (* finish : result -> taskState -> taskState *)
                   fun finish ans (ts as {tid, status, closure, out}) =
                       let
                         fun sameTid (NONE, NONE) = true
                           | sameTid (SOME t1, SOME t2) = CML.sameTid (t1,t2)
                           | sameTid _ = false
                       in
                         if not (sameTid (tid, SOME myTid)) then ts
                         else
                           (case status of
                              Running =>
                                (noteMsg ("  Completed execution and sending " ^
                                          "result ...");
                                 SyncVar.iPut (tValOf out, SUCCESS (tskId,ans));
                                 {tid=tid, status=Finished, closure=closure,
                                  out=out})
                            | _ => ts)
                       end
                 in 
                   (case (TaskState.updateStatus (tskId, updStat Waiting))
                         handle e => (TaskState.endTaskStart tskId; raise e) of
                      NONE   => ()
                    | SOME _ =>
                      let
                        val _ = TaskState.endTaskStart tskId

                        val _ = noteMsg ("  Initializing and waiting " ^
                                         "for input ...")
                        val env = SyncVar.iGet inp
                      in
                        (case (TaskState.updateStatus (tskId, updStat Running)) of
                           NONE => ()
                         | SOME _ =>
                           let
                             val _ = noteMsg ("  Received input and starting " ^
                                              "execution ...")
     
                             val ans = f env
                           in
                             TaskState.update (tskId, finish ans)
                           end)
                      end)
                 end
           in
             (CML.spawn thread; ())
           end

       val ({closure=(f,x), out, ...}) = TaskState.get tskId

       val input = SyncVar.iVar ()
       val _ = SyncVar.iPut (input, x)

       val output = if newIvar then SOME (SyncVar.iVar ()) else out
     in
       spawnCord f input output
     end

  (* enable : task -> unit

     Enable a disabled task to start.  Does nothing for a task with
     any other status.
  *)
  fun enable tskId =
      let
        val _ = noteMsg ("Enabling task " ^ (taskIdToString tskId) ^ " ...")
        
        val _ = TaskState.beginTaskStart tskId
        val s = (status tskId)
                handle e => (TaskState.endTaskStart tskId; raise e)
      in
        case s of
          Disabled => startTask false tskId
        | _ => TaskState.endTaskStart tskId
      end

  (* inject : bool -> (env -> result) * env -> task

     Register a task with the "network".
  *)
  fun inject bEnable tsk =
      let
        val taskId = (!newTaskId) ()

        val _ = noteMsg ("Register new task " ^ (taskIdToString taskId) ^
                         " ...")
        val _ = TaskState.add(taskId,
                  {tid=NONE, status=Disabled, closure=tsk,
                   out=(SOME (SyncVar.iVar ()))})
        val _ = dbgMsg ("Finished registering new task " ^
                        (taskIdToString taskId) ^ " ...")
      in
        if bEnable then (enable taskId; taskId)
        else taskId
      end

  (* recvResultEvt : task -> result event
  
     Get result of a task.
  *)
  fun recvResultEvt otherTskId =
      let
        val _ = noteMsg ("Asking for result from " ^
                         (taskIdToString otherTskId) ^ "...")

        fun loop () =
            let
              val ({out,...}) = TaskState.get otherTskId
            in
              case out of
                NONE => (CML.yield (); loop ())
              | SOME outvar => SyncVar.iGetEvt outvar
            end
      in
        loop ()
      end

  (* wrap : 'a event * ('a -> 'b) -> 'b event *)
  fun wrap (evt, f) =
      let
        fun wrapper (FAILURE (tskid,g)) = FAILURE (tskid, f o g)
          | wrapper (SUCCESS (tskid,x)) = SUCCESS (tskid, f x)
      in
        CML.wrap (evt, wrapper)
      end

  (* guard : (unit -> 'a event) -> 'a event *)
  val guard = CML.guard

  (* sync : 'a event -> 'a *)
  fun sync evt = 
      let
        fun restart fromTask = 
            let
              val _ = TaskState.beginTaskStart fromTask
              val s = (status fromTask)
                      handle e => (TaskState.endTaskStart fromTask; raise e)
            in
              case s of
                Failed =>
                  let
                    val _ = noteMsg ("Restarting failed task " ^
                                     (taskIdToString fromTask) ^ " ...")
                    val _ = TaskState.update (fromTask, 
                              fn {tid,status,closure,out} =>
                                {tid=tid, status=Disabled, closure=closure,
                                 out=out})
                  in
                    startTask true fromTask
                  end
              | _ => TaskState.endTaskStart fromTask
            end

        fun getResult () =
            let
              val a = CML.sync evt
            in
              case a of
                FAILURE (fromTask,f) =>
                  (restart fromTask;
                   sync (wrap (recvResultEvt fromTask, f)))
              | SUCCESS (fromTask,ans) => 
                let
                  val _ = noteMsg ("Received result from " ^
                                   (taskIdToString fromTask) ^ ".")
                in
                  ans
                end
            end

        exception DoesntMatter
        fun update () =
            let
              val t = (SOME (getTaskId ())) handle InvalidTask => NONE
            in
              (case t of
                 NONE => getResult ()
               | SOME tskId =>
                 (case (TaskState.updateStatus (tskId, updStat Waiting)) of
                    (* Task failed, just end *)
                    NONE => raise DoesntMatter
                  | SOME _ =>
                    let
                      val ans = getResult ()
                    in
                      (case (TaskState.updateStatus (tskId, updStat Running)) of
                        (* Task failed, just end *)
                         NONE => raise DoesntMatter
                       | SOME _ => ans)
                    end))
            end

      in
        update ()
      end

  (* choose : 'a event list -> 'a event *)
  val choose = CML.choose

  (* select : 'a event list -> 'a *)
  fun select events = sync (choose events)

  (* recvResult : task -> result *)
  val recvResult = sync o recvResultEvt

  (* forget : task -> unit *)
  fun forget tskId = 
      let
        fun upd (ts as {tid,status,closure,out}) =
            let
              fun cleanup () =
                  let
                    val _ =  SyncVar.iPut (tValOf out,
                                           FAILURE (tskId, fn x => x))
                  in
                    {tid=tid,status=Failed,closure=closure,out=NONE}
                  end
            in
              case status of
                Running => cleanup ()
              | Waiting => cleanup ()
              | _       => ts
            end
      in
        TaskState.update (tskId, upd)
      end

  (* exit : result -> 'a
  
     Exit the current task and send result.
  *)
  fun exit ans =
      let
        val tskId = getTaskId ()

        fun upd (ts as {tid,status,closure,out}) =
            (case status of
               Running =>
                  let
                    val _ = noteMsg "Exiting early and sending result ...";
                    val _ = SyncVar.iPut (tValOf out, SUCCESS (tskId,ans))
                  in
                    {tid=tid,status=Finished,closure=closure,out=out}
                  end
             | _ => ts)
      in
        (TaskState.update (tskId, upd);
         CML.exit ())
      end

  (* forgetall : unit -> unit

     Forget all remaining tasks
  *)
  fun forgetall () =
      let
	fun forgetall' n =
              (forget n;
               if n < TaskState.numTasks () then forgetall' (n+1) else ())
      in
        forgetall' 0
      end

  (* startSim : real -> (env -> result) * env -> result
     startSim failureParam tsk = r

     Start the simulation with the given initial task 'tsk'.
     The 'failureParam' [0.0,1.0] gives how often a task will fail;
     0.0 indicates that tasks never fail and 1.0 indicates a task will
     fail whenever given an opportunity.
  *)
  fun startSim failureParam tsk =
      let
        val _ = TextIO.print "Starting ConCert Simulator ...\n"

        (* failureSim : unit SyncVar.ivar -> unit

           A thread that is always running that can be
           used to simulate failures.
        *)
        fun failureSim stop =
            let
              fun maybeKill (NONE) = ()
                | maybeKill (SOME victim) =
                  let
                    val stat = status victim

                    fun doit () =
                        let
                          val _ = forget victim
                        in
                          (case (status victim) of
                             Failed => noteMsg ("Task " ^
                                                (taskIdToString victim) ^
                                                " failed!")
                           | _ => ())
                        end
                  in
                    (case stat of
                       Running => doit ()
                     | Waiting => doit ()
                     | _       => ())
                  end

              fun loop () =
                  (case SyncVar.iGetPoll stop of
                     NONE =>
                       (if willFail failureParam then maybeKill (randTask ())
                        else ();
                        CML.yield ();
                        loop ())
                   | SOME _ => ())
            in
              loop ()
            end

        val result = ref NONE

        fun startup () =
            let
              val _ = newTaskId := createIdSrc ()
              val _ = TaskState.init ()
              val stop = SyncVar.iVar ()
              val _ = CML.spawn (fn () => failureSim stop)
              val taskId = inject true tsk
            in
              (result := SOME (recvResult taskId);
               forgetall ();
               SyncVar.iPut (stop, ()))
            end

        val _ = RunCML.doit (startup, NONE)
      in
        valOf (!result)
      end

end; (* structure CCTasksR *)
