(****************************************************************************)
(* rwlocks.sml                                                              *)
(*                                                                          *)
(* A implementation of locks for concurrent reading of and exclusive        *)
(* writing to some state.                                                   *)
(*                                                                          *)
(* Based on code by Andrew Appel found at:                                  *)
(*   http://www.cs.princeton.edu/courses/archive/fall01/cs510/crxw/crxw.sml *)
(*                                                                          *)
(* $Id: rwlocks.sml,v 1.1.1.1 2002/05/15 22:00:06 bechang_conductor Exp $       *)
(****************************************************************************)


signature RWLOCKS =
sig

  type lock

  (* Create a new lock *)
  val lock : unit -> lock

  (* Reading *)
  val beginRead  : lock -> unit CML.event
  val endRead    : lock -> unit

  (* Writing *)
  val beginWrite : lock -> unit CML.event
  val endWrite   : lock -> unit

end; (* signature RWLOCKS *)


structure RWLocks :> RWLOCKS =
struct
  open CML

  (*******************************)
  (* Locks                       *)
  (*******************************)
  datatype lock = Lock of unit chan  (* begin reading *)
                        * unit chan  (* end reading *)
                        * unit chan  (* begin writing *)
                        * unit chan  (* end writing *)
                        * int  chan  (* request writing *)


  (*******************************)
  (* Implementation              *)
  (*******************************)

  (* lock : unit -> lock *)
  fun lock () =
      let
        val bRead    = channel ()
        val eRead    = channel ()
        val bWrite   = channel ()
        val eWrite   = channel ()
        val reqWrite = channel ()

        fun when x y = if x then y else never

        fun serverLoop (nreaders, nwrite_reqs) =
              select [
                (* A reader can begin only when there are no write
                   requests. *)
                when (nwrite_reqs=0)
                     (wrap (recvEvt bRead,
                            fn () => serverLoop (nreaders+1, nwrite_reqs))),

                (* When a reader ends, just decrement the number of
                   readers working. *)
                wrap (recvEvt eRead,
                      fn () => serverLoop (nreaders-1, nwrite_reqs)),

                (* A write can begin only when there are no readers working. *)
                when (nreaders=0)
                     (wrap (recvEvt bWrite,
                            fn () => writing (nwrite_reqs-1))),

                (* When a write request comes in, just add/remove it to/from
                   the queue *)
                wrap (recvEvt reqWrite,
                      fn i => serverLoop (nreaders, nwrite_reqs+i))
              ]

        and writing (nwrite_reqs) =
              select [
                (* When a writer finishes, restart the loop listening
                   for new requests *)
                wrap (recvEvt eWrite, fn () => serverLoop (0, nwrite_reqs)),

                (* When a write request comes in, just add/remove it to/from
                   the queue *)
                wrap (recvEvt reqWrite, fn i => writing (nwrite_reqs+i))
              ]

        (* Start the server thread *)
        val _ = spawn (fn () => serverLoop (0,0))
      in
        Lock (bRead, eRead, bWrite, eWrite, reqWrite)
      end

      (* beginRead : lock -> unit event *)
      fun beginRead (Lock (bRead, _, _, _, _)) = sendEvt (bRead, ())
    
      (* endRead : lock -> unit event *)
      fun endRead (Lock (_, eRead, _, _, _)) = send (eRead, ())
    
      (* beginWrite : lock -> unit event *)
      fun beginWrite (Lock (_, _, bWrite, _, reqWrite)) =
            withNack (fn nack =>
                         let
                           (* Request the write *)
                           val _ = send (reqWrite, 1)
    
                           (* If user thread decided not to actually write by
                              the time the write request could be satisfied,
                              tell the server that the user aborted. *)
                           val _ = spawn (fn () => (sync nack;
                                                    send (reqWrite, ~1)))
                         in
                           sendEvt (bWrite, ())
                         end)
    
      (* endWrite : lock -> unit event *)
      fun endWrite (Lock (_, _, _, eWrite, _)) = send (eWrite, ())

end; (* structure RWLocks *)
